import asyncio
import multiprocessing
import os
from concurrent.futures import ProcessPoolExecutor, ThreadPoolExecutor
from functools import partial
from typing import Any, Callable, Optional, Sequence

import cloudpickle
import psutil
from networkx import MultiDiGraph
from tenacity import Retrying, stop_after_attempt

from cubed.runtime.asyncio import async_map_dag
from cubed.runtime.backup import use_backups_default
from cubed.runtime.pipeline import visit_nodes
from cubed.runtime.types import Callback, CubedPipeline, DagExecutor, TaskEndEvent
from cubed.runtime.utils import (
    asyncio_run,
    execution_stats,
    execution_timing,
    handle_operation_start_callbacks,
    profile_memray,
)
from cubed.spec import Spec


def exec_stage_func(input, func=None, config=None, name=None, compute_id=None):
    return func(input, config=config)


class SingleThreadedExecutor(DagExecutor):
    """The default execution engine that runs tasks sequentially uses Python loops."""

    @property
    def name(self) -> str:
        return "single-threaded"

    def execute_dag(
        self,
        dag: MultiDiGraph,
        callbacks: Optional[Sequence[Callback]] = None,
        spec: Optional[Spec] = None,
        compute_id: Optional[str] = None,
        **kwargs,
    ) -> None:
        for name, node in visit_nodes(dag):
            handle_operation_start_callbacks(callbacks, name)
            pipeline: CubedPipeline = node["pipeline"]
            for m in pipeline.mappable:
                result = exec_stage_func(
                    m,
                    pipeline.function,
                    config=pipeline.config,
                    name=name,
                    compute_id=compute_id,
                )
                if callbacks is not None:
                    event = TaskEndEvent(name=name, result=result)
                    [callback.on_task_end(event) for callback in callbacks]


@execution_timing
def run_func_threads(input, func=None, config=None, name=None, compute_id=None):
    return func(input, config=config)


@profile_memray
@execution_stats
def run_func_processes(input, func=None, config=None, name=None, compute_id=None):
    return func(input, config=config)


def unpickle_and_call(f, inp, **kwargs):
    import cloudpickle

    f = cloudpickle.loads(f)
    inp = cloudpickle.loads(inp)
    kwargs = {k: cloudpickle.loads(v) for k, v in kwargs.items()}
    return f(inp, **kwargs)


def check_runtime_memory(spec, max_workers):
    allowed_mem = spec.allowed_mem if spec is not None else None
    total_mem = psutil.virtual_memory().total
    if allowed_mem is not None:
        if total_mem < allowed_mem * max_workers:
            raise ValueError(
                f"Total memory on machine ({total_mem}) is less than allowed_mem * max_workers ({allowed_mem} * {max_workers} = {allowed_mem * max_workers})"
            )


def threads_create_futures_func(
    concurrent_executor, function: Callable[..., Any], retries: int = 2
):
    if retries != 0:
        retryer = Retrying(reraise=True, stop=stop_after_attempt(retries + 1))
        function = partial(retryer, function)

    def create_futures_func(input, **kwargs):
        return [
            (
                i,
                asyncio.wrap_future(concurrent_executor.submit(function, i, **kwargs)),
            )
            for i in input
        ]

    return create_futures_func


class ThreadsExecutor(DagExecutor):
    """An execution engine that uses Python asyncio."""

    def __init__(self, **kwargs):
        self.kwargs = kwargs

        # Tell NumPy to use a single thread
        # from https://stackoverflow.com/questions/30791550/limit-number-of-threads-in-numpy
        os.environ["MKL_NUM_THREADS"] = "1"
        os.environ["NUMEXPR_NUM_THREADS"] = "1"
        os.environ["OMP_NUM_THREADS"] = "1"
        os.environ["VECLIB_MAXIMUM_THREADS"] = "1"

    @property
    def name(self) -> str:
        return "threads"

    def execute_dag(
        self,
        dag: MultiDiGraph,
        callbacks: Optional[Sequence[Callback]] = None,
        spec: Optional[Spec] = None,
        compute_id: Optional[str] = None,
        **kwargs,
    ) -> None:
        merged_kwargs = {**self.kwargs, **kwargs}
        asyncio_run(
            self._async_execute_dag(
                dag,
                callbacks=callbacks,
                spec=spec,
                compute_id=compute_id,
                **merged_kwargs,
            )
        )

    async def _async_execute_dag(
        self,
        dag: MultiDiGraph,
        callbacks: Optional[Sequence[Callback]] = None,
        spec: Optional[Spec] = None,
        compute_arrays_in_parallel: Optional[bool] = None,
        **kwargs,
    ) -> None:
        max_workers = kwargs.pop("max_workers", os.cpu_count())
        if spec is not None:
            check_runtime_memory(spec, max_workers)
            if "use_backups" not in kwargs and use_backups_default(spec):
                kwargs["use_backups"] = True

        concurrent_executor = ThreadPoolExecutor(max_workers=max_workers)
        try:
            create_futures_func = threads_create_futures_func(
                concurrent_executor, run_func_threads, kwargs.pop("retries", 2)
            )
            await async_map_dag(
                create_futures_func,
                dag=dag,
                callbacks=callbacks,
                compute_arrays_in_parallel=compute_arrays_in_parallel,
                **kwargs,
            )
        finally:
            # don't wait for any cancelled tasks
            concurrent_executor.shutdown(wait=False)


def processes_create_futures_func(concurrent_executor, function: Callable[..., Any]):
    def create_futures_func(input, **kwargs):
        # Pickle the function, args, and kwargs using cloudpickle.
        # They will be unpickled by unpickle_and_call.
        pickled_kwargs = {k: cloudpickle.dumps(v) for k, v in kwargs.items()}
        return [
            (
                i,
                asyncio.wrap_future(
                    concurrent_executor.submit(
                        unpickle_and_call,
                        cloudpickle.dumps(function),
                        cloudpickle.dumps(i),
                        **pickled_kwargs,
                    )
                ),
            )
            for i in input
        ]

    return create_futures_func


class ProcessesExecutor(DagExecutor):
    """An execution engine that uses local processes."""

    def __init__(self, **kwargs):
        self.kwargs = kwargs

        # Tell NumPy to use a single thread
        # from https://stackoverflow.com/questions/30791550/limit-number-of-threads-in-numpy
        os.environ["MKL_NUM_THREADS"] = "1"
        os.environ["NUMEXPR_NUM_THREADS"] = "1"
        os.environ["OMP_NUM_THREADS"] = "1"
        os.environ["VECLIB_MAXIMUM_THREADS"] = "1"

    @property
    def name(self) -> str:
        return "processes"

    def execute_dag(
        self,
        dag: MultiDiGraph,
        callbacks: Optional[Sequence[Callback]] = None,
        spec: Optional[Spec] = None,
        compute_id: Optional[str] = None,
        **kwargs,
    ) -> None:
        merged_kwargs = {**self.kwargs, **kwargs}
        asyncio_run(
            self._async_execute_dag(
                dag,
                callbacks=callbacks,
                spec=spec,
                compute_id=compute_id,
                **merged_kwargs,
            )
        )

    async def _async_execute_dag(
        self,
        dag: MultiDiGraph,
        callbacks: Optional[Sequence[Callback]] = None,
        spec: Optional[Spec] = None,
        compute_arrays_in_parallel: Optional[bool] = None,
        **kwargs,
    ) -> None:
        max_workers = kwargs.pop("max_workers", os.cpu_count())
        if spec is not None:
            check_runtime_memory(spec, max_workers)
            if "use_backups" not in kwargs and use_backups_default(spec):
                kwargs["use_backups"] = True

        use_processes = kwargs.pop("use_processes", True)
        if isinstance(use_processes, str):
            context = multiprocessing.get_context(use_processes)
        else:
            context = multiprocessing.get_context("spawn")

        # max_tasks_per_child is only supported from Python 3.11
        max_tasks_per_child = kwargs.pop("max_tasks_per_child", None)
        if max_tasks_per_child is None:
            concurrent_executor = ProcessPoolExecutor(
                max_workers=max_workers, mp_context=context
            )
        else:
            concurrent_executor = ProcessPoolExecutor(
                max_workers=max_workers,
                mp_context=context,
                max_tasks_per_child=max_tasks_per_child,
            )
        try:
            create_futures_func = processes_create_futures_func(
                concurrent_executor, run_func_processes
            )
            await async_map_dag(
                create_futures_func,
                dag=dag,
                callbacks=callbacks,
                compute_arrays_in_parallel=compute_arrays_in_parallel,
                **kwargs,
            )
        finally:
            # don't wait for any cancelled tasks
            concurrent_executor.shutdown(wait=False)



import collections
import copy
import logging
import time
from functools import partial
from typing import (
    Any,
    Callable,
    Dict,
    Iterable,
    Iterator,
    List,
    Optional,
    Sequence,
    Tuple,
    Union,
)

from lithops.executors import FunctionExecutor
from lithops.retries import RetryingFunctionExecutor, RetryingFuture
from lithops.wait import ALWAYS, ANY_COMPLETED
from networkx import MultiDiGraph

from cubed.runtime.backup import should_launch_backup, use_backups_default
from cubed.runtime.pipeline import visit_node_generations, visit_nodes
from cubed.runtime.types import Callback, DagExecutor
from cubed.runtime.utils import (
    handle_callbacks,
    handle_operation_start_callbacks,
    profile_memray,
)
from cubed.spec import Spec

logger = logging.getLogger(__name__)


@profile_memray
def run_func(input, func=None, config=None, name=None, compute_id=None):
    result = func(input, config=config)
    return result


def map_unordered(
    lithops_function_executor: RetryingFunctionExecutor,
    group_map_functions: Sequence[Callable[..., Any]],
    group_map_iterdata: Sequence[
        Iterable[Union[List[Any], Tuple[Any, ...], Dict[str, Any]]]
    ],
    group_names: Sequence[str],
    include_modules: Optional[List[str]] = None,
    timeout: Optional[int] = None,
    retries: int = 2,
    use_backups: bool = False,
    return_stats: bool = False,
    wait_dur_sec: Optional[int] = 1,
    **kwargs,
) -> Iterator[Any]:
    """
    Apply a function to items of an input list, yielding results as they are completed
    (which may be different to the input order).

    A generalisation of Lithops `map`, with retries, and relaxed return ordering.

    :param lithops_function_executor: The Lithops function executor to use.
    :param group_map_functions: A sequence of functions to map over the data.
    :param group_map_iterdata: A sequence of iterables of input data.
    :param group_names: The names of the function/iterable groups.
    :param include_modules: Modules to include.
    :param retries: The number of times to retry a failed task before raising an exception.
    :param use_backups: Whether to launch backup tasks to mitigate against slow-running tasks.
    :param return_stats: Whether to return lithops stats.
    :param wait_dur_sec: Time interval to wait between each check for completed tasks.

    :return: Function values (and optionally stats) as they are completed, not necessarily in the input order.
    """
    return_when = ALWAYS if use_backups else ANY_COMPLETED
    wait_dur_sec = wait_dur_sec or 1

    future_to_group_name: Dict[str, str] = {}
    group_name_to_function: Dict[str, Callable[..., Any]] = {}
    # backups are launched based on task start and end times for the group
    start_times: Dict[str, Dict[RetryingFuture, float]] = {}
    end_times: Dict[str, Dict[RetryingFuture, float]] = collections.defaultdict(dict)
    backups: Dict[RetryingFuture, RetryingFuture] = {}
    pending: List[RetryingFuture] = []
    group_name: str

    for map_function, map_iterdata, group_name in zip(
        group_map_functions, group_map_iterdata, group_names
    ):
        # can't use functools.partial here as we get an error in lithops
        # also, lithops extra_args doesn't work for this case
        partial_map_function = lambda x, map_function=map_function: map_function(
            x, **kwargs
        )
        group_name_to_function[group_name] = partial_map_function

        futures = lithops_function_executor.map(
            partial_map_function,
            list(map_iterdata),  # lithops requires a list
            timeout=timeout,
            include_modules=include_modules or [],
            retries=retries,
        )
        start_times[group_name] = {k: time.monotonic() for k in futures}
        future_to_group_name.update({k: group_name for k in futures})
        pending.extend(futures)

    while pending:
        finished, pending = lithops_function_executor.wait(
            pending,
            throw_except=False,
            return_when=return_when,
            show_progressbar=False,
            wait_dur_sec=wait_dur_sec,
        )
        for future in finished:
            if future.error:
                # if the task has a backup that is not done, or is done with no exception, then don't raise this exception
                backup = backups.get(future, None)
                if backup:
                    if not backup.done or not backup.error:
                        continue
                future.status(throw_except=True)
            group_name = future_to_group_name[future]  # type: ignore[assignment]
            end_times[group_name][future] = time.monotonic()
            if return_stats:
                yield future.result(), standardise_lithops_stats(group_name, future)
            else:
                yield future.result()

            # remove any backup task
            if use_backups:
                backup = backups.get(future, None)
                if backup:
                    if backup in pending:
                        pending.remove(backup)
                    del backups[future]
                    del backups[backup]
                    backup.cancel()

        if use_backups:
            now = time.monotonic()
            for future in copy.copy(pending):
                group_name = future_to_group_name[future]  # type: ignore[assignment]
                if future not in backups and should_launch_backup(
                    future, now, start_times[group_name], end_times[group_name]
                ):
                    input = future.input
                    logger.warning(f"Running backup task for {group_name} on {input}")
                    futures = lithops_function_executor.map(
                        group_name_to_function[group_name],
                        [input],
                        timeout=timeout,
                        include_modules=include_modules or [],
                        retries=0,  # don't retry backup tasks
                    )
                    start_times[group_name].update(
                        {k: time.monotonic() for k in futures}
                    )
                    future_to_group_name.update({k: group_name for k in futures})
                    pending.extend(futures)
                    backup = futures[0]
                    backups[future] = backup
                    backups[backup] = future
            time.sleep(wait_dur_sec)


def execute_dag(
    dag: MultiDiGraph,
    callbacks: Optional[Sequence[Callback]] = None,
    spec: Optional[Spec] = None,
    compute_arrays_in_parallel: Optional[bool] = None,
    **kwargs,
) -> None:
    use_backups = kwargs.pop("use_backups", use_backups_default(spec))
    wait_dur_sec = kwargs.pop("wait_dur_sec", None)
    compute_id = kwargs.pop("compute_id")
    allowed_mem = spec.allowed_mem if spec is not None else None
    function_executor = FunctionExecutor(**kwargs)
    runtime_memory_mb = function_executor.config[function_executor.backend].get(
        "runtime_memory", None
    )
    if runtime_memory_mb is not None and allowed_mem is not None:
        runtime_memory = runtime_memory_mb * 1_000_000
        if runtime_memory < allowed_mem:
            raise ValueError(
                f"Runtime memory ({runtime_memory}) is less than allowed_mem ({allowed_mem})"
            )
    with RetryingFunctionExecutor(function_executor) as executor:
        if not compute_arrays_in_parallel:
            for name, node in visit_nodes(dag):
                handle_operation_start_callbacks(callbacks, name)
                pipeline = node["pipeline"]
                for result, stats in map_unordered(
                    executor,
                    [run_func],
                    [pipeline.mappable],
                    [name],
                    use_backups=use_backups,
                    return_stats=True,
                    wait_dur_sec=wait_dur_sec,
                    # kwargs below
                    func=pipeline.function,
                    config=pipeline.config,
                    name=name,
                    compute_id=compute_id,
                ):
                    handle_callbacks(callbacks, result, stats)
        else:
            for gen in visit_node_generations(dag):
                group_map_functions = []
                group_map_iterdata = []
                group_names = []
                for name, node in gen:
                    pipeline = node["pipeline"]
                    f = partial(
                        run_func, func=pipeline.function, config=pipeline.config
                    )
                    group_map_functions.append(f)
                    group_map_iterdata.append(pipeline.mappable)
                    group_names.append(name)
                for name in group_names:
                    handle_operation_start_callbacks(callbacks, name)
                for result, stats in map_unordered(
                    executor,
                    group_map_functions,
                    group_map_iterdata,
                    group_names,
                    use_backups=use_backups,
                    return_stats=True,
                    wait_dur_sec=wait_dur_sec,
                    # TODO: other kwargs (func, config, name)
                    compute_id=compute_id,
                ):
                    handle_callbacks(callbacks, result, stats)


def standardise_lithops_stats(name: str, future: RetryingFuture) -> Dict[str, Any]:
    stats = future.stats
    return dict(
        name=name,
        task_create_tstamp=stats["host_job_create_tstamp"],
        function_start_tstamp=stats["worker_func_start_tstamp"],
        function_end_tstamp=stats["worker_func_end_tstamp"],
        task_result_tstamp=stats["host_status_done_tstamp"],
        peak_measured_mem_start=stats["worker_peak_memory_start"],
        peak_measured_mem_end=stats["worker_peak_memory_end"],
    )


class LithopsExecutor(DagExecutor):
    """An execution engine that uses Lithops."""

    def __init__(self, **kwargs):
        self.kwargs = kwargs

    @property
    def name(self) -> str:
        return "lithops"

    def execute_dag(
        self,
        dag: MultiDiGraph,
        callbacks: Optional[Sequence[Callback]] = None,
        spec: Optional[Spec] = None,
        compute_id: Optional[str] = None,
        **kwargs,
    ) -> None:
        merged_kwargs = {**self.kwargs, **kwargs}
        execute_dag(
            dag,
            callbacks=callbacks,
            spec=spec,
            compute_id=compute_id,
            **merged_kwargs,
        )



# Copied from numpy.array_api
import builtins

from cubed.array_api.inspection import __array_namespace_info__
from cubed.backend_array_api import namespace as nxp

int8 = nxp.int8
int16 = nxp.int16
int32 = nxp.int32
int64 = nxp.int64
uint8 = nxp.uint8
uint16 = nxp.uint16
uint32 = nxp.uint32
uint64 = nxp.uint64
float32 = nxp.float32
float64 = nxp.float64
complex64 = nxp.complex64
complex128 = nxp.complex128
bool = nxp.bool

_all_dtypes = (
    int8,
    int16,
    int32,
    int64,
    uint8,
    uint16,
    uint32,
    uint64,
    float32,
    float64,
    complex64,
    complex128,
    bool,
)
_boolean_dtypes = (bool,)
_real_floating_dtypes = (float32, float64)
_floating_dtypes = (float32, float64, complex64, complex128)
_complex_floating_dtypes = (complex64, complex128)
_integer_dtypes = (int8, int16, int32, int64, uint8, uint16, uint32, uint64)
_signed_integer_dtypes = (int8, int16, int32, int64)
_unsigned_integer_dtypes = (uint8, uint16, uint32, uint64)
_integer_or_boolean_dtypes = (
    bool,
    int8,
    int16,
    int32,
    int64,
    uint8,
    uint16,
    uint32,
    uint64,
)
_real_numeric_dtypes = (
    float32,
    float64,
    int8,
    int16,
    int32,
    int64,
    uint8,
    uint16,
    uint32,
    uint64,
)
_numeric_dtypes = (
    float32,
    float64,
    complex64,
    complex128,
    int8,
    int16,
    int32,
    int64,
    uint8,
    uint16,
    uint32,
    uint64,
)

_dtype_categories = {
    "all": _all_dtypes,
    "real numeric": _real_numeric_dtypes,
    "numeric": _numeric_dtypes,
    "integer": _integer_dtypes,
    "integer or boolean": _integer_or_boolean_dtypes,
    "boolean": _boolean_dtypes,
    "real floating-point": _floating_dtypes,
    "complex floating-point": _complex_floating_dtypes,
    "floating-point": _floating_dtypes,
}


# A Cubed-specific utility.
def _upcast_integral_dtypes(
    x, dtype=None, *, allowed_dtypes=("numeric",), fname=None, device=None
):
    """Ensure the input dtype is allowed. If it's None, provide a good default dtype."""
    dtypes = __array_namespace_info__().default_dtypes(device=device)

    # Validate.
    is_invalid = all(x.dtype not in _dtype_categories[a] for a in allowed_dtypes)
    if is_invalid:
        errmsg = f"Only {' or '.join(allowed_dtypes)} dtypes are allowed"
        if fname:
            errmsg += f" in {fname}"
        raise TypeError(errmsg)

    # Choose a good default dtype, when None
    if dtype is None:
        if x.dtype in _boolean_dtypes:
            dtype = dtypes["integral"]
        elif x.dtype in _signed_integer_dtypes:
            dtype = dtypes["integral"]
        elif x.dtype in _unsigned_integer_dtypes:
            # produce an unsigned integer dtype at the same default precision
            dtype = dtypes["integral"]
            if dtype == int8:
                dtype = uint8
            elif dtype == int16:
                dtype = uint16
            elif dtype == int32:
                dtype = uint32
            elif dtype == int64:
                dtype = uint64
        else:
            dtype = x.dtype

    return dtype


def _promote_scalars(x1, x2, op):
    """Promote at most one of x1 or x2 to an array from a Python scalar"""
    x1_is_scalar = isinstance(x1, (int, float, complex, builtins.bool))
    x2_is_scalar = isinstance(x2, (int, float, complex, builtins.bool))
    if x1_is_scalar and x2_is_scalar:
        raise TypeError(f"At least one of x1 and x2 must be an array in {op}")
    elif x1_is_scalar:
        x1 = x2._promote_scalar(x1)
    elif x2_is_scalar:
        x2 = x1._promote_scalar(x2)
    return x1, x2



import warnings
from collections import Counter

from cubed.runtime.pipeline import visit_nodes
from cubed.runtime.types import Callback


class MemoryWarningCallback(Callback):
    def on_compute_start(self, event):
        # store ops keyed by name
        self.ops = {}
        for name, node in visit_nodes(event.dag):
            primitive_op = node["primitive_op"]
            self.ops[name] = primitive_op

        # count number of times each op exceeds allowed mem
        self.counter = Counter()

    def on_task_end(self, event):
        allowed_mem = self.ops[event.name].allowed_mem
        if (
            event.peak_measured_mem_end is not None
            and event.peak_measured_mem_end > allowed_mem
        ):
            self.counter.update({event.name: 1})

    def on_compute_end(self, event):
        if sum(self.counter.values()) > 0:
            exceeded = [
                f"{k} ({v}/{self.ops[k].num_tasks})" for k, v in self.counter.items()
            ]
            warnings.warn(
                f"Peak memory usage exceeded allowed_mem when running tasks: {', '.join(exceeded)}",
                UserWarning,
                stacklevel=2,
            )



from typing import Tuple

from cubed.backend_array_api import namespace as nxp
from cubed.core.ops import map_selection
from cubed.types import T_RectangularChunks
from cubed.utils import _cumsum, normalize_chunks
from cubed.vendor.dask.array.overlap import coerce_boundary, coerce_depth
from cubed.vendor.dask.utils import has_keyword


def map_overlap(
    func,
    *args,
    dtype=None,
    chunks=None,
    depth=None,
    boundary=None,
    trim=False,
    **kwargs,
):
    """Apply a function to corresponding blocks from multiple input arrays with some overlap.

    Parameters
    ----------
    func : callable
        Function to apply to every block (with overlap) to produce the output array.
    args : arrays
        The Cubed arrays to map over. Note that currently only one array may be specified.
    dtype : np.dtype
        The ``dtype`` of the output array.
    chunks : tuple
        Chunk shape of blocks in the output array.
    depth :  int, tuple, dict or list
        The number of elements that each block should share with its neighbors.
    boundary : value type, tuple, dict or list
        How to handle the boundaries. Note that this currently only supports constant values.
    trim : bool
        Whether or not to trim ``depth`` elements from each block after calling the map function.
        Currently only ``False`` is supported.
    **kwargs : dict
        Extra keyword arguments to pass to function.
    """
    if trim:
        raise ValueError("trim is not supported")

    chunks = normalize_chunks(chunks, dtype=dtype)
    shape = tuple(map(sum, chunks))

    # Coerce depth and boundary arguments to lists of individual
    # specifications for each array argument
    def coerce(xs, arg, fn):
        if not isinstance(arg, list):
            arg = [arg] * len(xs)
        return [fn(x.ndim, a) for x, a in zip(xs, arg)]

    depth = coerce(args, depth, coerce_depth)
    boundary = coerce(args, boundary, coerce_boundary)
    new_axis = kwargs.pop("new_axis", None)
    if new_axis is not None:
        raise NotImplementedError("new_axis is not supported in map_overlap")

    x = args[0]  # TODO: support multiple input arrays

    def selection_function(out_key):
        out_coords = out_key[1:]
        block_id = out_coords
        return get_item_with_depth(x.chunks, block_id, depth[0])

    max_num_input_blocks = _overlap_num_input_blocks(depth[0], x.numblocks)

    has_block_id_kw = has_keyword(func, "block_id")

    # First read the chunk with overlaps determined by depth, then pad boundaries second.
    # Do it this way round so we can do everything with one blockwise. The alternative,
    # which pads the entire array first (via concatenate), would result in at least one extra copy.

    return map_selection(
        _overlap,
        selection_function,
        x,
        shape=shape,
        dtype=dtype,
        chunks=chunks,
        max_num_input_blocks=max_num_input_blocks,
        overlap_func=func,
        depth=depth,
        boundary=boundary,
        numblocks=x.numblocks,
        has_block_id_kw=has_block_id_kw,
        **kwargs,
    )


def _overlap_num_input_blocks(depth, numblocks):
    num = 1
    for i in depth.keys():
        num *= min(numblocks[i], 3)
    return num


def _overlap(
    a,
    overlap_func=None,
    depth=None,
    boundary=None,
    numblocks=None,
    has_block_id_kw=False,
    block_id=None,
    **kwargs,
):
    depth = depth[0]
    boundary = boundary[0]
    out = _pad_boundaries(a, depth, boundary, numblocks, block_id)
    if has_block_id_kw:
        return overlap_func(out, block_id=block_id, **kwargs)
    else:
        return overlap_func(out, **kwargs)


def _clamp(minimum: int, x: int, maximum: int) -> int:
    return max(minimum, min(x, maximum))


def get_item_with_depth(
    chunks: T_RectangularChunks, idx: Tuple[int, ...], depth
) -> Tuple[slice, ...]:
    """Convert a chunk index to a tuple of slices with depth offsets."""
    starts = tuple(_cumsum(c, initial_zero=True) for c in chunks)

    def depth_offsets(d):
        if isinstance(d, int):
            return -d, d
        return d

    loc = tuple(
        (
            _clamp(0, start[i] + depth_offsets(depth[ax])[0], start[-1]),
            _clamp(0, start[i + 1] + depth_offsets(depth[ax])[1], start[-1]),
        )
        for ax, (i, start) in enumerate(zip(idx, starts))
    )
    return tuple(slice(*s, None) for s in loc)


def _pad_boundaries(x, depth, boundary, numblocks, block_id):
    for i in range(x.ndim):
        d = depth.get(i, 0)
        if d == 0 or block_id[i] not in (0, numblocks[i] - 1) or boundary[i] == "none":
            continue
        pad_shape = list(x.shape)
        pad_shape[i] = d
        pad_shape = tuple(pad_shape)
        p = nxp.full_like(x, fill_value=boundary[i], shape=pad_shape)
        if block_id[i] == 0:  # first block on axis i
            x = nxp.concat([p, x], axis=i)
        if block_id[i] == numblocks[i] - 1:  # last block on axis i
            x = nxp.concat([x, p], axis=i)
    return x



import asyncio
from typing import Optional, Sequence

import ray
from networkx import MultiDiGraph

from cubed.runtime.asyncio import async_map_dag
from cubed.runtime.backup import use_backups_default
from cubed.runtime.types import Callback, DagExecutor
from cubed.runtime.utils import asyncio_run, execute_with_stats
from cubed.spec import Spec


class RayExecutor(DagExecutor):
    """An execution engine that uses Ray."""

    def __init__(self, **kwargs):
        self.kwargs = kwargs

    @property
    def name(self) -> str:
        return "ray"

    def execute_dag(
        self,
        dag: MultiDiGraph,
        callbacks: Optional[Sequence[Callback]] = None,
        spec: Optional[Spec] = None,
        compute_id: Optional[str] = None,
        **kwargs,
    ) -> None:
        merged_kwargs = {**self.kwargs, **kwargs}

        ray_init = merged_kwargs.pop("ray_init", None)
        if ray_init is not None:
            ray.init(**ray_init)

        asyncio_run(
            self._async_execute_dag(
                dag,
                callbacks=callbacks,
                spec=spec,
                compute_id=compute_id,
                **merged_kwargs,
            )
        )

    async def _async_execute_dag(
        self,
        dag: MultiDiGraph,
        callbacks: Optional[Sequence[Callback]] = None,
        spec: Optional[Spec] = None,
        compute_arrays_in_parallel: Optional[bool] = None,
        **kwargs,
    ) -> None:
        if spec is not None:
            if "use_backups" not in kwargs and use_backups_default(spec):
                kwargs["use_backups"] = True

        allowed_mem = spec.allowed_mem if spec is not None else 2_000_000_000
        retries = kwargs.pop("retries", 2)

        # note we can define the remote function here (doesn't need to be at top-level), and pass in memory and retries
        @ray.remote(memory=allowed_mem, max_retries=retries, retry_exceptions=True)
        def run_remotely(input, func=None, config=None, name=None, compute_id=None):
            # note we can't use the execution_stat decorator since it doesn't work with ray decorators
            result, stats = execute_with_stats(func, input, config=config)
            return result, stats

        def create_futures_func(input, **kwargs):
            return [
                (i, asyncio.wrap_future(run_remotely.remote(i, **kwargs).future()))
                for i in input
            ]

        await async_map_dag(
            create_futures_func,
            dag=dag,
            callbacks=callbacks,
            compute_arrays_in_parallel=compute_arrays_in_parallel,
            **kwargs,
        )
